# [1] will be removed at last
# [2] will be replaced by logging
# [3] convert to np.float32
# ---------------------------------
import sys  # [1]
sys.path.append('/data3/zrn/workspace/github') # [1]

import os
from glob import glob

# NumPy
import numpy as np
from astropy.stats import mad_std, sigma_clip
# AstroPy
from astropy.table import Table

from drpy import CCDDataList
from drpy.longslit import (response, illumination, concatenate, fitcoords, resample, 
                            trace, background, extract, dispcor)
from drpy.plotting import plotSpectrum1D, plot2d
from drpy.utils import imstatistics



def loadList(listname, listpath=''):
    """Load a file list.

    Parameters
    ----------
    listname : string
        Name of the list.
    listpath : string
        Path to the list.

    Returns
    -------
    filelist : list
        The loaded file list.
    """

    with open(os.path.join(listpath, f"{listname}.list"), 'r') as f:
        filelist = [line.strip() for line in f.readlines()]
    
    return filelist

def loadLists(listnames, listpath=''):
    """Load file lists.

    Parameters
    ----------
    listnames : list
        A list of list names.
    listpath : string
        Path to the lists.

    Returns
    -------
    filelists : dict
        Loaded lists stored in a dictionary.
    """
    
    lists = dict()
    for listname in listnames:
        lists[listname] = loadList(listname, listpath=listpath)
    
    return lists

def genFileTable(ccddatalist):
    """
    """
    
    file_table = list()
    for ccd in ccddatalist:
        file_table.append([
            ccd.header['FILENAME'], ccd.header['OBSTYPE'], ccd.header['OBJECT'], 
            ccd.header['EXPTIME']])
    file_table = Table(
        rows=file_table, names=['file_name', 'file_type', 'targ_name', 'exposure'])
    file_table.pprint_all() # [2]
    
    return file_table

def main(data_dir, save_dir, hdu, row_range, col_range, slit_along, rdnoise, gain):
    """
    """
    
    # ==================================================================================
    # Path
    # ==================================================================================
    plot_path = os.path.join(save_dir, 'figs')
    if not os.path.exists(plot_path):
        os.makedirs(plot_path)
    fits_path = os.path.join(save_dir, 'fits')
    if not os.path.exists(fits_path):
        os.makedirs(fits_path)
    
    # ==================================================================================
    # Load data
    # ==================================================================================
    file_list = sorted(glob(os.path.join(data_dir, '*.fit')))
    ccddatalist = CCDDataList.read(file_list=file_list, hdu=hdu)

    # ==================================================================================
    # Organize data
    # ==================================================================================
    file_table = genFileTable(ccddatalist)

    mask_bias = file_table['file_type'] == 'BIAS'
    mask_flat = file_table['file_type'] == 'SPECLFLAT'
    mask_lamp = file_table['file_type'] == 'SPECLLAMP'
    mask_spec = file_table['file_type'] == 'SPECLTARGET'
    
    bias_list = ccddatalist[mask_bias]
    flat_list = ccddatalist[mask_flat]
    lamp_list = ccddatalist[mask_lamp]
    spec_list = ccddatalist[mask_spec]

    # ==================================================================================
    # Trim
    # ==================================================================================
    bias_list_trimmed = bias_list.trim(row_range=row_range, col_range=col_range)
    flat_list_trimmed = flat_list.trim(row_range=row_range, col_range=col_range)
    lamp_list_trimmed = lamp_list.trim(row_range=row_range, col_range=col_range)
    spec_list_trimmed = spec_list.trim(row_range=row_range, col_range=col_range)

    # ==================================================================================
    # Bias combination
    # ==================================================================================
    master_bias = bias_list_trimmed.combine(
        method='average', sigma_clip=True, sigma_clip_low_thresh=3, 
        sigma_clip_high_thresh=3, sigma_clip_func=np.ma.median, 
        sigma_clip_dev_func=mad_std, mem_limit=500e6)

    # Check statistics 
    bias_list_trimmed.statistics(verbose=True) # [2]
    imstatistics(master_bias, verbose=True) # [2]

    # Plot master bias
    plot2d(master_bias.data, title='master bias', show=False, save=True, path=plot_path)

    # Write master bias to file
    master_bias.write(
        os.path.join(fits_path, 'master_bias.fits'), overwrite=True) # [3]
    
    # ==================================================================================
    # Bias subtraction
    # ==================================================================================
    # !!! Uncertainties assigned here (equal to that of ``master_bias``) is useless !!!
    flat_list_bias_subtracted = flat_list_trimmed - master_bias
    lamp_list_bias_subtracted = lamp_list_trimmed - master_bias
    spec_list_bias_subtracted = spec_list_trimmed - master_bias
    
    # Check statistics
    flat_list_trimmed.statistics(verbose=True) # [2]
    flat_list_bias_subtracted.statistics(verbose=True) # [2]
    
    # ==================================================================================
    # Flat Combination
    # ==================================================================================
    # !!! Old uncertainties are overwritten !!!
    combined_flat = flat_list_bias_subtracted.combine(
        method='average', sigma_clip=True, sigma_clip_low_thresh=3, 
        sigma_clip_high_thresh=3, sigma_clip_func=np.ma.median, 
        sigma_clip_dev_func=mad_std, mem_limit=500e6)

    # Check statistics
    imstatistics(combined_flat, verbose=True) # [2]

    # Plot combined flat
    plot2d(
        combined_flat.data, title='combined flat', show=False, save=True, path=plot_path)

    # Write combined flat to file
    combined_flat.write(
        os.path.join(fits_path, 'combined_flat.fits'), overwrite=True) # [3]
    
    # ==================================================================================
    # Flat normalization
    # ==================================================================================
    # Response calibration
    # [sometimes ``n_piece`` needs to be tuned.]
    reflat = response(
        ccd=combined_flat, slit_along=slit_along, n_piece=19, sigma_lower=3, 
        sigma_upper=3, grow=10, use_mask=True, show=False, save=True, path=plot_path)

    # Check statistics
    imstatistics(reflat, verbose=True) # [2]

    # Plot response calibrated flat
    plot2d(reflat.data, title='reflat', show=False, save=True, path=plot_path)

#     # Custom mask
#     reflat.mask[520:620, 1200:1750] = True
    
    # Plot response mask
    plot2d(
        reflat.mask.astype(int), vmin=0, vmax=1, title='response mask', show=False, 
        save=True, path=plot_path)

    # Write response calibrated flat to file
    reflat.write(os.path.join(fits_path, 'reflat.fits'), overwrite=True) # [3]
    
    # Illumination modeling
#     ilflat = illumination(
#         ccd=reflat, slit_along=slit_along, method='iraf',
#         bins=[0, 250, 500, 750, 1000, 1100, 1200, 1250, 1300, 1350, 1400, 1450, 1500, 
#               1550, 1600, 1650, 1700, 1750, 1800, 1850, 1900, 1950], n_piece=[3, 3, 3, 
#         3, 12, 15, 15, 15, 15, 21, 15, 15, 12, 17, 21, 23, 25, 45, 29, 30, 29], 
#         n_iter=5, sigma_lower=1.5, sigma_upper=3, grow=3, use_mask=True, show=False, 
#         save=True, path=plot_path)

#     ilflat = illumination(
#         ccd=reflat, slit_along=slit_along, method='iraf',
#         bins=[0, 250, 500, 750, 1000, 1100, 1200, 1250, 1300, 1350, 1400, 1450, 1500, 
#               1550, 1600, 1650, 1700, 1750, 1800, 1850, 1900, 1950], n_piece=[4, 4, 4, 
#         4, 15, 15, 15, 15, 15, 21, 15, 12, 10, 12, 21, 23, 23, 16, 20, 16, 25], 
#         n_iter=3, sigma_lower=1.5, sigma_upper=3, grow=3, use_mask=True, show=False, 
#         save=True, path=plot_path)

#     ilflat = illumination(
#         ccd=reflat, slit_along=slit_along, method='CubicSpline2D', n_piece=(21, 5), 
#         n_iter=3, sigma_lower=1.5, sigma_upper=3, grow=5, use_mask=True, show=False, 
#         save=True, path=plot_path)
    
    ilflat = illumination(
        ccd=reflat, slit_along=slit_along, method='Gaussian2D', sigma=(20, 20), bins=10, 
        n_iter=10, sigma_lower=1.5, sigma_upper=3, grow=5, use_mask=True, show=False, 
        save=True, path=plot_path)

    # Plot illumination
    plot2d(ilflat.data, title='illumination', show=False, save=True, path=plot_path)
    
    # Plot illumination mask
    plot2d(
        ilflat.mask.astype(int), vmin=0, vmax=1, title='illumination mask', show=False, 
        save=True, path=plot_path)
    
    # Write illumination to file
    ilflat.write(os.path.join(fits_path, 'illumination.fits'), overwrite=True) # [3]
    
    # Normalization
    normalized_flat = reflat.divide(ilflat, handle_meta='first_found')

    # Plot normalized flat
    plot2d(
        normalized_flat.data, title='normalized flat', show=False, save=True, 
        path=plot_path)

    # Write normalized flat to file
    normalized_flat.write(
        os.path.join(fits_path, 'normalized_flat.fits'), overwrite=True)  # [3]
    
    # [can be replaced by a pre-defined custom mask]
    normalized_flat.mask = None

    # ==================================================================================
    # Create deviation
    # ==================================================================================
    # !!! generate real uncertainty here!!!
    lamp_list_bias_subtracted_with_deviation = (
        lamp_list_bias_subtracted.create_deviation(gain=gain, readnoise=rdnoise)
    )
    spec_list_bias_subtracted_with_deviation = (
        spec_list_bias_subtracted.create_deviation(gain=gain, readnoise=rdnoise)
    )
    
    # ==================================================================================
    # Flat-fielding
    # ==================================================================================
    # Lamp
    lamp_list_flat_corrected = (
        lamp_list_bias_subtracted_with_deviation / normalized_flat
    )
    # Spec
    spec_list_flat_corrected = (
        spec_list_bias_subtracted_with_deviation / normalized_flat
    )

    # ==================================================================================
    # Cosmic ray removal
    # ==================================================================================
    spec_list_cosmicray_corrected = spec_list_flat_corrected.cosmicray(
        method='Laplacian', use_mask=False, gain=gain, readnoise=rdnoise, sigclip=4.5, 
        sigfrac=0.3, objlim=1, niter=5, verbose=True)

    # Plot cosmic-ray removed frames [how to output all the images???]
    plot2d(
        spec_list_cosmicray_corrected[0].data, title='cosmic-ray removed', show=False, 
        save=True, path=plot_path)

    # ==================================================================================
    # Concatenate lamp
    # ==================================================================================
    # Ensure that the first is the short exposure 
    if np.diff(file_table[mask_lamp]['exposure'].data)[0] < 0:
        lamp_list_flat_corrected = lamp_list_flat_corrected[::-1]

    concatenated_lamp = concatenate(
        lamp_list_flat_corrected, slit_along=slit_along, index=665)

    # Plot concatenated lamp
    plot2d(
        concatenated_lamp.data, title='concatenated lamp', show=False, save=True, 
        path=plot_path)

    # Write concatenated lamp to file
    concatenated_lamp.write(
        os.path.join(fits_path, 'concatenated_lamp.fits'), overwrite=True)  # [3]

    # ==================================================================================
    # Fit coordinates
    # ==================================================================================
    U, V = fitcoords(
        ccd=concatenated_lamp, slit_along=slit_along, method='2D', n_med=5, n_ext=10, 
        n_sub=5, n_piece=(3, 3), n_iter=3, sigma_lower=3, sigma_upper=3, grow=False, 
        height=0, threshold=0, distance=5, prominence=0.001, width=5, wlen=10, 
        rel_height=1, plateau_size=1, show=False, save=True, path=plot_path)
    
    # ==================================================================================
    # Distortion correction
    # ==================================================================================
    # Lamp
    resampled_lamp = resample(
        ccd=concatenated_lamp, slit_along=slit_along, U=U, V=V, use_uncertainty=True, 
        verbose=False)
    
    # Plot resampled lamp
    plot2d(
        resampled_lamp.data, title='resampled lamp', show=False, save=True, 
        path=plot_path)
    
    # Write resampled lamp to file
    resampled_lamp.write(
        os.path.join(fits_path, 'resampled_lamp.fits'), overwrite=True) # [3]

    # Spec
    for i, spec_cosmicray_corrected in enumerate(spec_list_cosmicray_corrected):
        
        old_name = spec_cosmicray_corrected.header['FILENAME']
        new_name = '{}_corrected.fits'.format(old_name[:-4])
        
        print('{} -> {}'.format(old_name, new_name))
        
        resampled_spec = resample(
            ccd=spec_cosmicray_corrected, slit_along=slit_along, U=U, V=V, 
            use_uncertainty=True, verbose=False)

        # Plot resampled spectrum
        plot2d(
            resampled_spec.data, title=old_name, show=False, save=True, 
            path=plot_path)

        # Write resampled standard spectrum to file
        resampled_spec.write(
            os.path.join(fits_path, new_name), overwrite=True) # [3]

    # ==================================================================================
    # Wavelength calibration
    # ==================================================================================
    lamp1d = extract(
        ccd=resampled_lamp, slit_along=slit_along, trace=750, n_aper=1, aper_width=10, 
        use_uncertainty=True, show=False, save=True, path=plot_path)
    
    calibrated_lamp1d = dispcor(
        spectrum1d=lamp1d, reverse=True, file_name='bfosc_slit23_g4.fits', n_ext=10, 
        n_sub=5, n_piece=3, refit=True, n_iter=5, sigma_lower=3, sigma_upper=3, 
        grow=False, show=False, save=True, path=plot_path)
    
    # Plot
    plotSpectrum1D(
        spectrum1d=calibrated_lamp1d, xlabel='wavelength [A]', ylabel='count', 
        title='extracted lamp', show=False, save=True, path=plot_path)
    
    calibrated_lamp1d.write(
        os.path.join(fits_path, 'lamp1d.fits'), overwrite=True)

if __name__ == '__main__':

    # Path
    data_dir = '/data3/zrn/workspace/data_reduction/demo_SN2021klg/data'
    save_dir = '/data3/zrn/workspace/data_reduction/demo_SN2021klg'
#     data_dir = '/data3/zrn/workspace/data_reduction/demo_standard/data'
#     save_dir = '/data3/zrn/workspace/data_reduction/demo_standard'
    
    # Hyper parameters
    hdu = 0
    col_range = (0, 1950)
    row_range = (329, 1830)
    slit_along = 'col'
    rdnoise = 4.64
    gain = 1.41

    main(
        data_dir=data_dir, save_dir=save_dir, hdu=hdu, row_range=row_range, 
        col_range=col_range, slit_along=slit_along, rdnoise=rdnoise, gain=gain)



try:
    from prettytable import PrettyTable
    from prettytable import HEADER, NONE
except ImportError:
    print( 'Module `prettytable` not found. Please install with: pip install prettytable' )
    sys.exit()

try:
    import matplotlib.pyplot as plt
    # Set plot parameters
    plt.rcParams['axes.linewidth'] = 2.0
    plt.rcParams['mathtext.fontset'] = 'stix'
    plt.rcParams['font.family'] = 'STIXGeneral'
except ImportError:
    print( 'Module `matplotlib` not found. Please install with: pip install matplotlib' )
    sys.exit()

def print_statistics( imglist ):
    '''
    '''
    tab = PrettyTable( hrules = HEADER, vrules = NONE )
    tab.field_names = [ 'FRAME TYPE', 'SHAPE', 'MEAN', 'STDDEV', 'MIN', 'MAX' ]
    for img in imglist:
        tab.add_row([ 'Image',
                       img.data.shape,
                       round( img.data.mean(), 2 ),
                       round( img.data.std( ddof = 1 ), 2 ),
                       int( img.data.min() ),
                       int( img.data.max() ) ])
        tab.add_row([ 'Uncertainty',
                      '\"',
                      round( img.uncertainty.array.mean(), 2 ),
                      round( img.uncertainty.array.std( ddof = 1 ), 2 ),
                      round( img.uncertainty.array.min(), 2 ),
                      round( img.uncertainty.array.max(), 2 ) ])
    for col in ['MEAN', 'STDDEV', 'MIN', 'MAX']: tab.align[col] = 'r'
    print( '\n' + tab.get_string() + '\n' )

def plot2d( img, title, show = 1, save = 0 ):
    '''
    '''
    
    fig, ax = plt.subplots( 1, 1, figsize = ( 8, 8 ) )
    fig.subplots_adjust( right = 0.8 )
    
    # Image
    im = ax.imshow( img.data, cmap = 'Greys_r', origin = 'lower', extent = ( 0.5, img.data.shape[1]+0.5, 0.5, img.data.shape[0]+0.5) )

    # Colorbar
    cax = fig.add_axes([ ax.get_position().x1 + 0.02, ax.get_position().y0, 0.04, ax.get_position().height ])
    cb = fig.colorbar( im, cax = cax )

    # Settings
    ax.tick_params( which = 'major', direction = 'in', color = 'w', top = True, right = True, length = 7, width = 1.5, labelsize = 18 )
    ax.set_xlabel( 'X', fontsize = 22 ); ax.set_ylabel( 'Y', fontsize = 22 )
    cb.ax.tick_params( which = 'major', direction = 'in', color = 'w', right = True, length = 7, width = 1.5, labelsize = 18 )
    cb.ax.set_ylabel( 'ADU', fontsize = 22 )
    ax.set_title( title, fontsize = 22 )

    if save:
        fig_path = 'figs'
        if not os.path.exists( fig_path ): os.makedirs( fig_path )
        print( F'[Plotting] Save to { os.path.join( fig_path, F"{title}.png".replace( " ", "_" ) ) }' )
        plt.savefig( os.path.join( fig_path, F'{title}.png'.replace( ' ', '_' ) ), dpi = 144 )

    if show:
        print( '[Plotting] Show plot' )
        plt.show()
    plt.close()


# [1] will be removed at last
# [2] will be replaced by logging
# [3] convert to np.float32
# ---------------------------------
import sys  # [1]
sys.path.append('/home/zrn/Workspace/') # [1]

import os
from glob import glob

# NumPy
import numpy as np
from astropy.stats import mad_std, sigma_clip
# AstroPy
from astropy.table import Table

from pyloss import CCDDataList
from pyloss.longslit import (response, illumination, concatenate, fitcoords, resample, 
                             trace, background, extract)
from pyloss.plotting import plot2d
from pyloss.utils import imstatistics



def pipeline(
        data_dir, save_dir, hdu, arc_name, std_name, obj_name, row_range, col_range, 
        slit_along, rdnoise, gain):
    """
    """
    
    # ==================================================================================
    # Path
    # ==================================================================================
    plot_path = os.path.join(save_dir, 'figs')
    if not os.path.exists(plot_path):
        os.makedirs(plot_path)
    fits_path = os.path.join(save_dir, 'fits')
    if not os.path.exists(fits_path):
        os.makedirs(fits_path)
    
    # ==================================================================================
    # Load data
    # ==================================================================================
    file_list = sorted(glob(os.path.join(data_dir, '*.fit')))
    ccddatalist = CCDDataList.read(file_list=file_list, hdu=hdu)

    # ==================================================================================
    # Organize data
    # ==================================================================================
    file_table = list()
    for ccd in ccddatalist:
        file_table.append([
            ccd.header['FILENAME'], ccd.header['OBSTYPE'], ccd.header['OBJECT'], 
            ccd.header['EXPTIME']])
    file_table = Table(
        rows=file_table, names=['file_name', 'file_type', 'targ_name', 'exposure'])
    file_table.pprint_all() # [2]
    
    mask_slittarget = file_table['file_type'] == 'SLITTARGET'
    mask_bias = (~mask_slittarget) & (file_table['targ_name'] == 'BIAS')
    mask_flat = (~mask_slittarget) & (file_table['targ_name'] == 'FLAT')
    mask_arc = (~mask_slittarget) & (file_table['targ_name'] == arc_name)
    mask_std = (~mask_slittarget) & (file_table['targ_name'] == std_name)
    mask_obj = (~mask_slittarget) & (file_table['targ_name'] == obj_name)
    
    slittargetlist = ccddatalist[mask_slittarget]
    biaslist = ccddatalist[mask_bias]
    flatlist = ccddatalist[mask_flat]
    arclist = ccddatalist[mask_arc]
    stdlist = ccddatalist[mask_std]
    objlist = ccddatalist[mask_obj]

    # ==================================================================================
    # Trim
    # ==================================================================================
    biaslist_trimmed = biaslist.trim(row_range=row_range, 
                                     col_range=col_range)
    flatlist_trimmed = flatlist.trim(row_range=row_range, 
                                     col_range=col_range)
    arclist_trimmed = arclist.trim(row_range=row_range, 
                                   col_range=col_range)
    stdlist_trimmed = stdlist.trim(row_range=row_range, 
                                   col_range=col_range)
    objlist_trimmed = objlist.trim(row_range=row_range, 
                                   col_range=col_range)

    # ==================================================================================
    # Bias combination
    # ==================================================================================
    master_bias = biaslist_trimmed.combine(
        method='average', sigma_clip=True, sigma_clip_low_thresh=3, 
        sigma_clip_high_thresh=3, sigma_clip_func=np.ma.median, 
        sigma_clip_dev_func=mad_std, mem_limit=500e6)

    # Check statistics 
    biaslist_trimmed.statistics(verbose=True) # [2]
    imstatistics(master_bias, verbose=True) # [2]

    # Plot master bias
    plot2d(master_bias.data, title='master bias', save=True, path=plot_path)

    # Write master bias to file
    master_bias.write(
        os.path.join(fits_path, 'master_bias.fits'), overwrite=True) # [3]
    
    # ==================================================================================
    # Bias subtraction
    # ==================================================================================
    # !!! Uncertainties assigned here (equal to that of ``master_bias``) is useless !!!
    # Flat
    flatlist_bias_subtracted = flatlist_trimmed - master_bias
    
    # Arc
    arclist_bias_subtracted = arclist_trimmed - master_bias
    
    # Standand
    stdlist_bias_subtracted = stdlist_trimmed - master_bias
    
    # Object
    objlist_bias_subtracted = objlist_trimmed - master_bias
    
    # Check statistics
    flatlist_trimmed.statistics(verbose=True) # [2]
    flatlist_bias_subtracted.statistics(verbose=True) # [2]
    
    # ==================================================================================
    # Flat Combination
    # ==================================================================================
    # !!! Old uncertainties are overwritten !!!
    combined_flat = flatlist_bias_subtracted.combine(
        method='average', sigma_clip=True, sigma_clip_low_thresh=3, 
        sigma_clip_high_thresh=3, sigma_clip_func=np.ma.median, 
        sigma_clip_dev_func=mad_std, mem_limit=500e6)

    # Check statistics
    imstatistics(combined_flat, verbose=True) # [2]

    # Plot combined flat
    plot2d(combined_flat.data, title='combined flat', save=True, path=plot_path)

    # Write combined flat to file
    combined_flat.write(
        os.path.join(fits_path, 'combined_flat.fits'), overwrite=True) # [3]
    
    # ==================================================================================
    # Flat normalization
    # ==================================================================================
    # Response calibration
    reflat = response(
        ccd=combined_flat, slit_along=slit_along, n_piece=19, sigma_lower=3, 
        sigma_upper=3, grow=10, use_mask=True, save=True, path=plot_path)

    # Check statistics
    imstatistics(reflat, verbose=True) # [2]

    # Plot response calibrated flat
    plot2d(reflat.data, title='reflat', save=True, path=plot_path)

    # Plot response mask
    plot2d(
        reflat.mask.astype(int), vmin=0, vmax=1, title='response mask', save=True, 
        path=plot_path)

    # Write response calibrated flat to file
    reflat.write(os.path.join(fits_path, 'reflat.fits'), overwrite=True) # [3]

    # Illumination modeling
#     ilflat = illumination(
#         ccd=reflat, slit_along=slit_along, method='iraf',
#         bins=[0, 250, 500, 750, 1000, 1100, 1200, 1250, 1300, 1350, 1400, 1450, 1500, 
#               1550, 1600, 1650, 1700, 1750, 1800, 1850, 1900, 1950], n_piece=[4, 4, 4, 
#         4, 15, 15, 15, 15, 15, 21, 15, 12, 10, 12, 21, 23, 23, 16, 20, 16, 25], 
#         n_iter=3, sigma_lower=0.75, sigma_upper=3, grow=3, use_mask=True, show=False, 
#         save=True, path=plot_path)

#     ilflat = illumination(
#         ccd=reflat, slit_along=slit_along, method='CubicSpline2D', n_piece=(21, 5), 
#         n_iter=3, sigma_lower=1.5, sigma_upper=3, grow=5, use_mask=True, show=False, 
#         save=True, path=plot_path)
    
    ilflat = illumination(
        ccd=reflat, slit_along=slit_along, method='Gaussian2D', sigma=(50, 20), 
        n_iter=3, sigma_lower=1.5, sigma_upper=3, grow=5, use_mask=True, show=False, 
        save=True, path=plot_path)

    # Plot illumination
    plot2d(ilflat.data, title='illumination', save=True, path=plot_path)
    
    # Plot illumination mask
    plot2d(
        ilflat.mask.astype(int), vmin=0, vmax=1, title='illumination mask', save=True, 
        path=plot_path)
    
    # Output here???
    
    # Normalization
    normalized_flat = reflat.divide(ilflat, handle_meta='first_found')

    # Plot normalized flat
    plot2d(normalized_flat.data, title='normalized flat', save=True, path=plot_path)

    # Write normalized flat to file
    normalized_flat.write(
        os.path.join(fits_path, 'normalized_flat.fits'), overwrite=True)  # [3]
    
    # [can be replaced by a pre-defined custom mask]
    normalized_flat.mask = None

    # ==================================================================================
    # Create deviation
    # ==================================================================================
    # !!! generate real uncertainty here!!!
    # Arc
    arclist_bias_subtracted_with_deviation = arclist_bias_subtracted.create_deviation(
        gain=gain, readnoise=rdnoise)

    # Standand
    stdlist_bias_subtracted_with_deviation = stdlist_bias_subtracted.create_deviation(
        gain=gain, readnoise=rdnoise)

    # Object
    objlist_bias_subtracted_with_deviation = objlist_bias_subtracted.create_deviation(
        gain=gain, readnoise=rdnoise)
    
    # ==================================================================================
    # Flat-fielding
    # ==================================================================================
    # Arc
    arclist_flat_corrected = arclist_bias_subtracted_with_deviation / normalized_flat
    
    # Standard
    stdlist_flat_corrected = stdlist_bias_subtracted_with_deviation / normalized_flat
    
    # Object
    objlist_flat_corrected = objlist_bias_subtracted_with_deviation / normalized_flat

    # Output here???

    # ==================================================================================
    # Cosmic ray removal
    # ==================================================================================
    # Standard
    stdlist_cosmicray_corrected = stdlist_flat_corrected.cosmicray(
        method='Laplacian', use_mask=False, gain=gain, readnoise=rdnoise, sigclip=4.5, 
        sigfrac=0.3, objlim=1, niter=5, verbose=True)
    
    # Object
    objlist_cosmicray_corrected = objlist_flat_corrected.cosmicray(
        method='Laplacian', use_mask=False, gain=gain, readnoise=rdnoise, sigclip=4.5, 
        sigfrac=0.3, objlim=1, niter=5, verbose=True)

    # Plot cosmic-ray removed frames
    plot2d(
        stdlist_cosmicray_corrected[0].data, title='cosmic-ray removed standard', 
        save=True, path=plot_path)
    plot2d(
        objlist_cosmicray_corrected[0].data, title='cosmic-ray removed object', 
        save=True, path=plot_path)

    # Output here???

    # ==================================================================================
    # Concatenate arc spectrum
    # ==================================================================================
    concatenated_arc = concatenate(
        arclist_flat_corrected, slit_along=slit_along, index=665) # order???

    # Plot concatenated arc spectrum
    plot2d(concatenated_arc.data, title='concatenated arc', save=True, path=plot_path)

#     # ==================================================================================
#     # Fit coordinates
#     # ==================================================================================
#     X = fitcoords(
#         ccd=concatenated_arc, slit_along=slit_along, method='2D', n_med=5, n_ext=10, 
#         n_sub=5, n_piece=(3, 3), n_iter=3, sigma_lower=3, sigma_upper=3, grow=False, 
#         height=0, threshold=0, distance=5, prominence=0.001, width=5, wlen=10, 
#         rel_height=1, plateau_size=1, use_mask=True, show=False, save=True, 
#         path=plot_path)
    
#     # ==================================================================================
#     # Distortion correction
#     # ==================================================================================
#     # Arc
#     resampled_arc = resample(
#         ccd=concatenated_arc, slit_along=slit_along, X=X, use_uncertainty=True, 
#         verbose=False)
    
#     # Plot resampled arc spectrum
#     plot2d(resampled_arc.data, title='resampled arc', save=True, path=plot_path)
    
#     # Write resampled arc spectrum to file
#     resampled_arc.write(
#         os.path.join(fits_path, 'resampled_arc.fits'), overwrite=True) # [3]

#     # Standard
#     resampled_std = resample(
#         ccd=stdlist_cosmicray_corrected[0], slit_along=slit_along, X=X, 
#         use_uncertainty=True, verbose=False)
    
#     # Plot resampled standard spectrum
#     plot2d(
#         resampled_std.data, title='resampled standard', save=True, 
#         path=plot_path)
    
#     # Write resampled standard spectrum to file
#     resampled_std.write(
#         os.path.join(fits_path, 'resampled_standard.fits'), overwrite=True) # [3]
    
#     # Object???

#     # ==================================================================================
#     # Wavelength calibration
#     # ==================================================================================

    
#     # ==================================================================================
#     # Trace
#     # ==================================================================================
    trace_std = trace(
        ccd=stdlist_cosmicray_corrected[0], slit_along=slit_along, seeing=8, n_med=10, n_piece=3, 
        n_iter=5, sigma_lower=2, sigma_upper=2, grow=False, show=False, save=True, 
        path=plot_path)

    extract(
        ccd=concatenated_arc, slit_along=slit_along, trace=trace_std, n_aper=10, aper_width=400, 
        use_uncertainty=False, show=False, save=True, path=plot_path)
#     background_std = background(
#         ccd=resampled_std, slit_along=slit_along, trace=trace_std, distance=100, 
#         aper_width=100, degree=1, n_iter=5, sigma_lower=3, sigma_upper=3, grow=False)
#     # Plot background spectrum
#     plot2d(background_std.data, title='background', show=False, save=True, path=plot_path)
#     # Write background spectrum to file
#     background_std.write(
#         os.path.join(fits_path, 'background_std.fits'), overwrite=True) # [3]

#     # Background subtraction
#     std_background_subtracted = resampled_std.subtract(
#         background_std, handle_meta='first_found')
#     # Plot background subtracted standard spectrum
#     plot2d(
#         std_background_subtracted.data, title='background subtracted standard spectrum', 
#         show=False, save=True, path=plot_path)
#     # Write background subtracted standard spectrum to file
#     std_background_subtracted.write(
#         os.path.join(fits_path, 'background_subtracted_std.fits'), overwrite=True) # [3]

if __name__ == '__main__':

    # Parameters
    data_dir = '/home/zrn/Data_Reduction/20210504'
    std_name = 'Feige34'
    obj_name = '2021klg'
    
    # Hyper parameters
    hdu = 0
    arc_name = 'FeAr'
    col_range = (0, 1950)
    row_range = (329, 1830)
    slit_along = 'col'
    rdnoise = 4.64
    gain = 1.41

    pipeline(data_dir=data_dir, 
             save_dir=data_dir,
             hdu=hdu,
             arc_name=arc_name,
             std_name=std_name,
             obj_name=obj_name,
             row_range=row_range,
             col_range=col_range,
             slit_along=slit_along,
             rdnoise=rdnoise,
             gain=gain)



import os

import pyregion
# drpy
from drpy import __version__ as version_drpy

__all__ = ['MODULE_PATH', 'loadList', 'loadLists']

LIBRARY_PATH = os.path.join(os.path.split(__file__)[0], 'lib')


def login(grism):
    """Print login message."""
    
    print('===========================================================================')
    print(f'              Data Reduction Pipeline for 2.16-m/BFOSC {grism}')
    print(f'                         (based on drpy v{version_drpy})')
    print('---------------------------------------------------------------------------')
    
    
def loadList(list_name, list_path=''):
    """Load a file list."""

    with open(os.path.join(list_path, f'{list_name}.list'), 'r') as f:
        file_list = [line.strip() for line in f.readlines()]
    
    return file_list


def loadLists(list_names, list_path=''):
    """Load file lists."""
    
    list_dict = dict()
    for list_name in list_names:
        list_dict[list_name] = loadList(list_name, list_path=list_path)
    
    return list_dict


def getMask(region_name, shape):
    """Generate custom mask from ds9 region."""
    
    r = pyregion.open(region_name)
    
    mask = r.get_mask(shape=shape)
    
    return mask


"""
mini-Sitian pipeline (version 2023)
"""

import os
from glob import glob
from copy import deepcopy

# NumPy
import numpy as np
# AstroPy
from astropy.io import fits
from astropy.time import Time
from astropy.stats import mad_std
from astropy.nddata import CCDData
# ccdproc
from ccdproc import ImageFileCollection
# drpy
from drpy.batch import CCDDataList
from drpy.plotting import plot2d
from drpy.utils import imstatistics

from .utils import makeDirectory, modifyHeader


def pipeline(save_dir, data_dir, keywords, telescope, target, gain, rdnoise, row_range, 
             col_range, steps, mem_limit, verbose, show, save):
    """
    """

    # Make subdirectories
    fig_path = makeDirectory(parent=save_dir, child=f'{target}/fig')
    pro_path = makeDirectory(parent=save_dir, child=f'{target}/pro')
    cal_path = makeDirectory(parent=save_dir, child=f'{target}/cal')
    
    if 'header' in steps:
        
        ifc = ImageFileCollection(
            location=data_dir, keywords=keywords, find_fits_by_reading=False, 
            filenames=None, glob_include=f'{telescope}*.fit', glob_exclude=None, 
            ext=0)
        for file_name in ifc.files_filtered(include_path=True):
            modifyHeader(file_name)
    
    # Construct image file collection
    ifc = ImageFileCollection(
        location=data_dir, keywords=keywords, find_fits_by_reading=False, 
        filenames=None, glob_include=f'{telescope}*.fit', glob_exclude=None, 
        ext=0)
    
    # Bias
    if 'bias' in steps:
        
        # Filter
        ifc_bias = ifc.filter(
            regex_match=True, file=f'{telescope}-bias-\d{{8}}-\d{{4}}.fit')
        if verbose:
            ifc_bias.summary.pprint_all()
        
        # Load data
        bias_list = CCDDataList.read(
            file_list=ifc_bias.files_filtered(include_path=True), hdu=0)
        
        # Trim
        bias_list_trimmed = bias_list.trim(row_range=row_range, col_range=col_range)
        
        # Combine bias
        master_bias = bias_list_trimmed.combine(
            method='average', sigma_clip=True, sigma_clip_low_thresh=3, 
            sigma_clip_high_thresh=3, sigma_clip_func=np.ma.median, 
            sigma_clip_dev_func=mad_std, mem_limit=mem_limit, 
            output_file=os.path.join(cal_path, f'{telescope}_master_bias.fits'), 
            overwrite_output=True)

        # Check statistics 
        bias_list_trimmed.statistics(verbose=verbose)
        imstatistics(master_bias, verbose=verbose)

        # Plot master bias
        plot2d(
            master_bias.data, title=f'{telescope} master bias', show=show, save=save, 
            path=fig_path)
        
        del bias_list, bias_list_trimmed, master_bias
    
    # Flat
    if 'flat' in steps:
        
        # Filter
        ifc_flat = ifc.filter(
            regex_match=True, file=f'{telescope}-flat-\d{{8}}-\d{{4}}.fit')
        if verbose:
            ifc_flat.summary.pprint_all()
        
        # Load data
        flat_list = CCDDataList.read(
            file_list=ifc_flat.files_filtered(include_path=True), hdu=0)
        
        # Trim
        flat_list_trimmed = flat_list.trim(row_range=row_range, col_range=col_range)
        
        # Subtract bias
        master_bias = CCDData.read(
            os.path.join(cal_path, f'{telescope}_master_bias.fits'))
        flat_list_bias_subtracted = flat_list_trimmed - master_bias
        
        # Combine flat
        scaling_func = lambda ccd: 1 / np.ma.average(ccd)
        master_flat = flat_list_bias_subtracted.combine(
            method='average', scale=scaling_func, sigma_clip=True, 
            sigma_clip_low_thresh=3, sigma_clip_high_thresh=3, 
            sigma_clip_func=np.ma.median, sigma_clip_dev_func=mad_std, 
            mem_limit=mem_limit, 
            output_file=os.path.join(cal_path, f'{telescope}_master_flat.fits'), 
            overwrite_output=True)

        # Check statistics
        flat_list_trimmed.statistics(verbose=verbose)
        flat_list_bias_subtracted.statistics(verbose=verbose)
        imstatistics(master_flat, verbose=verbose)
        
        # Plot master flat
        plot2d(
            master_flat.data, title=f'{telescope} master flat', show=show, save=save, 
            path=fig_path)
        
        del (flat_list, flat_list_trimmed, master_bias, flat_list_bias_subtracted, 
             master_flat)
        
    # Target
    if 'target' in steps:

        # Filter
        ifc_targ = ifc.filter(
            regex_match=True, file=f'{telescope}-{target}-\d{{8}}-\d{{4}}.fit')
        if verbose:
            ifc_targ.summary.pprint_all()

        # Load data
        targ_list = CCDDataList.read(
            file_list=ifc_targ.files_filtered(include_path=True), hdu=0)

        # Trim
        targ_list_trimmed = targ_list.trim(row_range=row_range, col_range=col_range)

        # Subtract bias
        master_bias = CCDData.read(
            os.path.join(cal_path, f'{telescope}_master_bias.fits'))
        targ_list_bias_subtracted = targ_list_trimmed - master_bias

        # Generate real uncertainty
        targ_list_bias_subtracted_with_deviation = (
            targ_list_bias_subtracted.create_deviation(
                gain=gain, readnoise=rdnoise, disregard_nan=True)
        )
        
        # Flat-fielding
        master_flat = CCDData.read(
            os.path.join(cal_path, f'{telescope}_master_flat.fits'))
        targ_list_flat_fielded = targ_list_bias_subtracted_with_deviation / master_flat
        
        # Remove cosmic-ray
        targ_list_corrected = targ_list_flat_fielded.cosmicray(
            method='Laplacian', use_mask=False, gain=gain, readnoise=rdnoise, 
            sigclip=4.5, sigfrac=0.3, objlim=1, niter=5, verbose=verbose)

        # Write images to file
        for ccd in targ_list_corrected:    
            fits_name = '{}_corrected.fits'.format(ccd.header['FILENAME'][:-4])
            ccd.write(
                os.path.join(pro_path, fits_name), overwrite=True)
    
    
if __name__ == '__main__':
    
    save_dir = '/data3/zrn/workspace/data_reduction/20230124'
    data_dir = '/data3/zrn/data/mst/20230124'
    keywords = [
        'DATE-OBJ', 'OBJECT', 'OBJCTRA', 'OBJCTDEC', 'AIRMASS', 'EXPTIME', 'EGAIN']
    telescopes = ['mst2']
    target = 'C2022E3'
    gain = 0.24665763974189758
    rdnoise = 6.4
    row_range = (2250, 4250)
    col_range = (3750, 5750)
    steps = ['header', 'bias', 'flat', 'target']
    mem_limit = 16e9
    verbose = True
    show = False
    save = True
    
    for telescope in telescopes:
        pipeline(
            save_dir=save_dir, data_dir=os.path.join(data_dir, telescope), 
            keywords=keywords, telescope=telescope, target=target, gain=gain, 
            rdnoise=rdnoise, row_range=row_range, col_range=col_range, steps=steps, 
            mem_limit=mem_limit, verbose=verbose, show=show, save=save)



import os

# AstroPy
from astropy.io import fits
from astropy.time import Time


def makeDirectory(parent, child, verbose=False):
    """Make directory."""
    
    path = os.path.abspath(os.path.join(parent, child))
    
    if verbose:
        print(f'  - Making {path}...')
        
    os.makedirs(path, exist_ok=True)
    
    return path


def modifyHeader(file_name, verbose=False):
    """Modify fits header."""
    
    with fits.open(file_name, 'update') as f:
        
        for hdu in f:
            
            # `RADECSYS` -> `RADESYSa`
            if ('RADECSYS' in hdu.header) & ('RADESYSa' not in hdu.header):
                
                if verbose:
                    print(f'{file_name}: rename `RADECSYS` to `RADESYSa`.')
                
                hdu.header['RADESYSa'] = (
                    hdu.header['RADECSYS'], hdu.header.comments['RADECSYS'])
                del hdu.header['RADECSYS']
            
            # Set `MJD-OBS` from `DATE-OBS`
            if ('DATE-OBS' in hdu.header) & ('MJD-OBS' not in hdu.header):
                
                if verbose:
                    print(f'{file_name}: set `MJD-OBS` from `DATE-OBS`.')
                    
                hdu.header['MJD-OBS'] = (
                    Time(hdu.header['DATE-OBS'], format='fits').mjd, 
                    'Set from DATE-OBS'
                )


"""
2.16-m/BFOSC pipeline (version 2020)
"""

import sys
sys.path.append('/data3/zrn/workspace/github')

import os
from glob import glob

# NumPy
import numpy as np
from astropy.stats import mad_std
# AstroPy
from astropy.table import Table

from drpy import CCDDataList, concatenate, transform
from drpy.photometry.utils import getFWHM
from drpy.longslit import (response, illumination, align, fitcoords, dispcor, trace, 
                           background, extract)
from drpy.plotting import plotSpectrum1D, plot2d
from drpy.utils import imstatistics, invertCoordinateMap


def loadList(listname, listpath=''):
    """Load a file list.

    Parameters
    ----------
    listname : string
        Name of the list.

    listpath : string
        Path to the list.

    Returns
    -------
    filelist : list
        The loaded file list.
    """

    with open(os.path.join(listpath, f"{listname}.list"), 'r') as f:
        filelist = [line.strip() for line in f.readlines()]
    
    return filelist


def loadLists(listnames, listpath=''):
    """Load file lists.

    Parameters
    ----------
    listnames : list
        A list of list names.

    listpath : string
        Path to the lists.

    Returns
    -------
    filelists : dict
        Loaded lists stored in a dictionary.
    """
    
    lists = dict()
    for listname in listnames:
        lists[listname] = loadList(listname, listpath=listpath)
    
    return lists


def genFileTable(ccddatalist, listpath, verbose=False):
    """Generate file table."""
    
    lists = loadLists(listnames=['bias', 'flat', 'lamp', 'targ'], listpath=listpath)
    
    file_table = list()
    for ccd in ccddatalist:
        if ccd.header['FILENAME'] in lists['bias']:
            obstype = 'BIAS'
            obj = 'BIAS'
        elif ccd.header['FILENAME'] in lists['flat']:
            obstype = 'SPECLFLAT'
            obj = 'FLAT'
        elif ccd.header['FILENAME'] in lists['lamp']:
            obstype = 'SPECLLAMP'
            obj = 'FeAr'
        elif ccd.header['FILENAME'] in lists['targ']:
            obstype = 'SPECLTARGET'
            obj = 'TARGET'
        else:
            obstype = 'OTHER'
            obj = 'OTHER'
        file_table.append([
            ccd.header['FILENAME'], obstype, obj, ccd.header['EXPTIME']])

    file_table = Table(
        rows=file_table, names=['file_name', 'file_type', 'targ_name', 'exposure'])

    if verbose: file_table.pprint_all() # [2]
    
    return file_table


def main(data_dir, save_dir, hdu, row_range, col_range, slit_along, re_n_piece, 
         il_bins, il_n_piece, rdnoise, gain, index_concatenate, index_standard, 
         file_name, show, save, verbose):
    """2.16-m/BFOSC pipeline (version 2020)"""

    # ==================================================================================
    # Path
    # ==================================================================================
    plot_path = os.path.join(save_dir, 'figs')
    if not os.path.exists(plot_path):
        os.makedirs(plot_path)
    fits_path = os.path.join(save_dir, 'fits')
    if not os.path.exists(fits_path):
        os.makedirs(fits_path)
    
    # ==================================================================================
    # Load data
    # ==================================================================================
    file_list = sorted(glob(os.path.join(data_dir, '*.fit')))
    ccddatalist = CCDDataList.read(file_list=file_list, hdu=hdu)

    # ==================================================================================
    # Organize data
    # ==================================================================================
    file_table = genFileTable(ccddatalist, save_dir, verbose)

    mask_bias = file_table['file_type'] == 'BIAS'
    mask_flat = file_table['file_type'] == 'SPECLFLAT'
    mask_lamp = file_table['file_type'] == 'SPECLLAMP'
    mask_slit = file_table['file_type'] == 'OTHER'
    mask_spec = file_table['file_type'] == 'SPECLTARGET'
    
    bias_list = ccddatalist[mask_bias]
    flat_list = ccddatalist[mask_flat]
    lamp_list = ccddatalist[mask_lamp]
    slit_list = ccddatalist[mask_slit]
    spec_list = ccddatalist[mask_spec]

    # ==================================================================================
    # Trim
    # ==================================================================================
    if verbose: print('TRIM')

    bias_list_trimmed = bias_list.trim(row_range=row_range, col_range=col_range)
    flat_list_trimmed = flat_list.trim(row_range=row_range, col_range=col_range)
    lamp_list_trimmed = lamp_list.trim(row_range=row_range, col_range=col_range)
    spec_list_trimmed = spec_list.trim(row_range=row_range, col_range=col_range)

    # ==================================================================================
    # Bias combination
    # ==================================================================================
    if verbose: print('BIAS COMBINATION')

    master_bias = bias_list_trimmed.combine(
        method='average', sigma_clip=True, sigma_clip_low_thresh=3, 
        sigma_clip_high_thresh=3, sigma_clip_func=np.ma.median, 
        sigma_clip_dev_func=mad_std, mem_limit=500e6)

    # Check statistics 
    bias_list_trimmed.statistics(verbose=verbose)
    imstatistics(master_bias, verbose=verbose)

    # Plot master bias
    plot2d(master_bias.data, title='master bias', show=show, save=save, path=plot_path)

    # Write master bias to file
    master_bias.write(
        os.path.join(fits_path, 'master_bias.fits'), overwrite=True)

    # ==================================================================================
    # Bias subtraction
    # ==================================================================================
    if verbose: print('BIAS SUBTRACTION')

    # !!! Uncertainties assigned here (equal to that of ``master_bias``) are useless !!!
    flat_list_bias_subtracted = flat_list_trimmed - master_bias
    lamp_list_bias_subtracted = lamp_list_trimmed - master_bias
    spec_list_bias_subtracted = spec_list_trimmed - master_bias

    # Check statistics
    flat_list_trimmed.statistics(verbose=verbose) # [2]
    flat_list_bias_subtracted.statistics(verbose=verbose) # [2]

    # ==================================================================================
    # Flat Combination
    # ==================================================================================
    if verbose: print('FLAT COMBINATION')

    # !!! Old uncertainties are overwritten !!!
    combined_flat = flat_list_bias_subtracted.combine(
        method='average', sigma_clip=True, sigma_clip_low_thresh=3, 
        sigma_clip_high_thresh=3, sigma_clip_func=np.ma.median, 
        sigma_clip_dev_func=mad_std, mem_limit=500e6)

    # Check statistics
    imstatistics(combined_flat, verbose=verbose) # [2]

    # Plot combined flat
    plot2d(
        combined_flat.data, title='combined flat', show=show, save=save, path=plot_path)

    # Write combined flat to file
    combined_flat.write(
        os.path.join(fits_path, 'combined_flat.fits'), overwrite=True)
    
    # ==================================================================================
    # Flat normalization
    # ==================================================================================
    if verbose: print('FLAT NORMALIZATION')

    # Response calibration
    reflat = response(
        ccd=combined_flat, slit_along=slit_along, n_piece=re_n_piece, n_iter=5, 
        sigma_lower=3, sigma_upper=3, grow=10, use_mask=True, show=show, save=save, 
        path=plot_path)

    # Check statistics
    imstatistics(reflat, verbose=verbose) # [2]

    # Plot response calibrated flat
    plot2d(reflat.data, title='reflat', show=show, save=save, path=plot_path)
    
#     # Custom mask
#     reflat.mask[520:620, 1200:1750] = True
    
    # Plot response mask
    plot2d(
        reflat.mask.astype(int), vmin=0, vmax=1, title='response mask', show=show, 
        save=save, path=plot_path)

    # Write response calibrated flat to file
    reflat.write(os.path.join(fits_path, 'reflat.fits'), overwrite=True) # [3]
    
    # Illumination modeling
    ilflat = illumination(
        ccd=reflat, slit_along=slit_along, method='iraf', bins=il_bins, 
        n_piece=il_n_piece, n_iter=5, sigma_lower=1.5, sigma_upper=3, grow=3, 
        use_mask=True, show=show, save=save, path=plot_path)

    # Plot illumination
    plot2d(ilflat.data, title='illumination', show=show, save=save, path=plot_path)
    
    # Plot illumination mask
    plot2d(
        ilflat.mask.astype(int), vmin=0, vmax=1, title='illumination mask', show=show, 
        save=save, path=plot_path)
    
    # Write illumination to file
    ilflat.write(os.path.join(fits_path, 'illumination.fits'), overwrite=True) # [3]
    
    # Normalization
    normalized_flat = reflat.divide(ilflat, handle_meta='first_found')

    # Plot normalized flat
    plot2d(
        normalized_flat.data, title='normalized flat', show=show, save=save, 
        path=plot_path)

    # Write normalized flat to file
    normalized_flat.write(
        os.path.join(fits_path, 'normalized_flat.fits'), overwrite=True)  # [3]
    
    # [can be replaced by a pre-defined custom mask]
    normalized_flat.mask = None

    # ==================================================================================
    # Create deviation
    # ==================================================================================
    if verbose: print('CREATE DEVIATION')

    # !!! generate real uncertainty here!!!
    lamp_list_bias_subtracted_with_deviation = (
        lamp_list_bias_subtracted.create_deviation(
            gain=gain, readnoise=rdnoise, disregard_nan=True)
    )
    spec_list_bias_subtracted_with_deviation = (
        spec_list_bias_subtracted.create_deviation(
            gain=gain, readnoise=rdnoise, disregard_nan=True)
    )

    # ==================================================================================
    # Flat-fielding
    # ==================================================================================
    if verbose: print('FLAT-FIELDING')

    spec_list_flat_corrected = (
        spec_list_bias_subtracted_with_deviation / normalized_flat
    )

    # ==================================================================================
    # Cosmic ray removal
    # ==================================================================================
    if verbose: print('COSMIC RAY REMOVAL')

    spec_list_cosmicray_corrected = spec_list_flat_corrected.cosmicray(
        method='Laplacian', use_mask=False, gain=gain, readnoise=rdnoise, sigclip=4.5, 
        sigfrac=0.3, objlim=1, niter=5, verbose=True)

    # ==================================================================================
    # Concatenate lamp
    # ==================================================================================
    if verbose: print('CONCATENATE LAMP')

    if slit_along == 'col':
        row_range = (0, None); col_range = (0, index_concatenate)
    else:
        row_range = (0, index_concatenate); col_range = (0, None)

    scale = 1 / file_table[mask_lamp]['exposure'].data

    # Ensure that the first is the short exposure
    if scale[0] < scale[1]:
        lamp_list_bias_subtracted_with_deviation = (
            lamp_list_bias_subtracted_with_deviation[::-1]
        )
        scale = scale[::-1]

    concatenated_lamp = concatenate(
        lamp_list_bias_subtracted_with_deviation, row_range, col_range, scale=None)

    # Plot concatenated lamp
    plot2d(
        concatenated_lamp.data, title='concatenated lamp', show=show, save=save, 
        path=plot_path)

    # Write concatenated lamp to file
    concatenated_lamp.write(
        os.path.join(fits_path, 'concatenated_lamp.fits'), overwrite=True)  # [3]

    # ==================================================================================
    # Fit coordinates
    # ==================================================================================
    if verbose: print('FIT COORDINATES')

    U, _ = fitcoords(
        ccd=concatenated_lamp, slit_along=slit_along, order=1, n_med=15, n_piece=3, 
        prominence=1e-3, n_iter=3, sigma_lower=3, sigma_upper=3, grow=False, 
        use_mask=False, show=show, save=save, path=plot_path, height=0, threshold=0, 
        distance=5, width=5, wlen=15, rel_height=1, plateau_size=1)
    
    X, Y = invertCoordinateMap(U=U, V=None)

    # ==================================================================================
    # Distortion correction
    # ==================================================================================
    if verbose: print('DISTORTION CORRECTION')

    # Lamp
    transformed_lamp = transform(
        ccd=concatenated_lamp, X=X, Y=Y)

    # Plot transformed lamp
    plot2d(
        transformed_lamp.data, title='transformed lamp', show=show, save=save, 
        path=plot_path)
    
    # Write transformed lamp to file
    transformed_lamp.write(
        os.path.join(fits_path, 'transformed_lamp.fits'), overwrite=True) # [3]

    # Spectra
    target_list_corrected = list()
    for i, spec_cosmicray_corrected in enumerate(spec_list_cosmicray_corrected):

        old_name = spec_cosmicray_corrected.header['FILENAME'][:-4]

        if verbose: print('{} corrected.'.format(old_name))

        transformed_spec = transform(
            ccd=spec_cosmicray_corrected, X=X, Y=Y)
        
        transformed_spec.mask = None

        # Plot transformed spectrum
        plot2d(
            transformed_spec.data, title=old_name, show=show, save=save, 
            path=plot_path)

        target_list_corrected.append(transformed_spec)

    # ==================================================================================
    # Standard
    # ==================================================================================
    if verbose: print('OUTPUT STANDARD')

    standard = target_list_corrected[index_standard]

    # Plot standard
    plot2d(
        standard.data, title='corrected standard', show=show, save=save, path=plot_path)
    
    # Write standard to file
    standard.write(os.path.join(fits_path, 'standard.fits'), overwrite=True) # [3]

    # ==================================================================================
    # Target
    # ==================================================================================
    if verbose: print('OUTPUT TARGET')

    del target_list_corrected[index_standard]

    target_list_aligned = align(target_list_corrected, slit_along=slit_along)
    
    target = target_list_aligned.combine(
        method='average', sigma_clip=True, sigma_clip_low_thresh=3, 
        sigma_clip_high_thresh=3, sigma_clip_func=np.ma.median, 
        sigma_clip_dev_func=mad_std, mem_limit=500e6)

    # Plot target
    plot2d(
        target.data, title='corrected target', show=show, save=save, path=plot_path)
    
    # Write standard to file
    target.write(os.path.join(fits_path, 'target.fits'), overwrite=True) # [3]
    
    # ==================================================================================
    # Wavelength calibration
    # ==================================================================================
    if verbose: print('WAVELENGTH CALIBRATION')

    lamp1d = extract(
        ccd=transformed_lamp, slit_along=slit_along, trace=750, n_aper=1, 
        aper_width=10, show=show, save=save, path=plot_path)
    
    calibrated_lamp1d = dispcor(
        spectrum1d=lamp1d, reverse=True, file_name=file_name, n_piece=3, refit=True, 
        n_iter=5, sigma_lower=3, sigma_upper=3, grow=False, use_mask=True, show=show, 
        save=save, path=plot_path)
    
    # Plot calibrated arc spectrum
    plotSpectrum1D(
        spectrum1d=calibrated_lamp1d, xlabel='wavelength [A]', ylabel='count', 
        title='calibrated lamp', show=show, save=save, path=plot_path)
    
    # Write calibrated arc spectrum to file
    calibrated_lamp1d.write(
        os.path.join(fits_path, 'lamp1d.fits'), overwrite=True) # [3]
    

if __name__ == '__main__':

    # Path
    data_dir = '/data3/zrn/data/216/LRS/20200406'
    save_dir = '/data3/zrn/workspace/data_reduction/20200406'
#     data_dir = '/data3/zrn/data/216/LRS/20200413'
#     save_dir = '/data3/zrn/workspace/data_reduction/20200413'
#     data_dir = '/data3/zrn/data/216/LRS/20200420'
#     save_dir = '/data3/zrn/workspace/data_reduction/20200420'
#     data_dir = '/data3/zrn/data/216/LRS/20200423'
#     save_dir = '/data3/zrn/workspace/data_reduction/20200423'
    
    # Hyper parameters
    hdu = 0
    col_range = (0, 1950)
    row_range = (329, 1830)
    slit_along = 'col'
    response_dict = {
        'n_piece': 19
    }
    illumination_dict = {
        'bins': [0, 250, 500, 750, 1000, 1100, 1200, 1250, 1300, 1350, 1400, 1450, 
                 1500, 1550, 1600, 1650, 1700, 1750, 1800, 1850, 1900, 1950], 
        'n_piece': [3, 3, 3, 3, 12, 15, 15, 15, 15, 21, 15, 15, 12, 17, 21, 23, 25, 45, 
                    29, 30, 29]
    }
    rdnoise = 4.64
    gain = 1.41
    index_concatenate = 665
    index_standard = -1
    file_name = 'bfosc_slit23_g4_2020.fits'

    main(
        data_dir=data_dir, save_dir=save_dir, hdu=hdu, row_range=row_range, 
        col_range=col_range, slit_along=slit_along, 
        re_n_piece=response_dict['n_piece'], il_bins=illumination_dict['bins'], 
        il_n_piece=illumination_dict['n_piece'], rdnoise=rdnoise, 
        gain=gain, index_concatenate=index_concatenate, index_standard=index_standard, 
        file_name=file_name, show=False, save=True, verbose=True)



"""
BFOSC/G4 pipeline
"""

import os, argparse
from copy import deepcopy

# NumPy
import numpy as np
# AstroPy
from astropy.stats import mad_std
from astropy.nddata import CCDData
# ccdproc
from ccdproc import ImageFileCollection
# drpy
from drpy.batch import CCDDataList
from drpy.image import concatenate
from drpy.utils import imstatistics
from drpy.plotting import plot2d, plotSpectrum1D
from drpy.twodspec.longslit import (response, illumination, fitcoords, transform, 
                                    extract)
from drpy.twodspec.utils import invertCoordinateMap
from drpy.onedspec import dispcor, saveSpectrum1D

from ..utils import makeDirectory, modifyHeader
from .utils import LIBRARY_PATH, login, loadLists, getMask


def pipeline(save_dir, data_dir, hdu, keywords, steps, row_range, col_range, 
             slit_along, n_piece, sigma, index, rdnoise, gain, custom_mask, reference, 
             mem_limit, show, save, verbose, mode): 
    """BFOSC/G4 pipeline."""
    
    # Login message
    if verbose:
        login('Grism 4')
    
    # Make directories
    if verbose:
        print('\n[MAKE DIRECTORIES]')
        print(f'  - Changing working directory to {save_dir}...')
    os.chdir(save_dir)
    fig_path = makeDirectory(parent='', child='fig', verbose=verbose)
    pro_path = makeDirectory(parent='', child='pro', verbose=verbose)
    cal_path = makeDirectory(parent='', child='cal', verbose=verbose)
    
    # Load lists
    list_dict = loadLists(list_names=['lamp', 'targ'])
    
    # Construct image file collection
    ifc = ImageFileCollection(
        location=data_dir, keywords=keywords, find_fits_by_reading=False, 
        filenames=None, glob_include='*.fit', glob_exclude=None, ext=hdu)

    # Modify fits header
    #   Note that image file collection is constructed before header modification.
    if 'header' in steps:
        if verbose:
            print('\n[HEADER MODIFICATION]')
        for file_name in ifc.files_filtered(include_path=True):
            modifyHeader(file_name, verbose=verbose)
    
    if 'trim' in steps:
        custom_mask = custom_mask[row_range[0]:row_range[1], col_range[0]:col_range[1]]
        trim = True
    else:
        trim = False
    
    # Bias combination
    if ('bias.combine' in steps) or ('bias' in steps):
        
        if verbose:
            print('\n[BIAS COMBINATION]')
        
        if verbose:
            print('  - Loading bias...')
        # Filter
        ifc_bias = ifc.filter(regex_match=True, imagetyp='Bias Frame')
        # Load bias
        bias_list = CCDDataList.read(
            file_list=ifc_bias.files_filtered(include_path=True), hdu=hdu)
        
        # Trim
        if trim:
            if verbose:
                print('  - Trimming...')
            bias_list = bias_list.trim(row_range=row_range, col_range=col_range)
        
        bias_list.statistics(verbose=verbose)

        # Combine bias
        if verbose:
            print('  - Combining...')
        master_bias = bias_list.combine(
            method='average', sigma_clip=True, sigma_clip_low_thresh=3, 
            sigma_clip_high_thresh=3, sigma_clip_func=np.ma.median, 
            sigma_clip_dev_func=mad_std, mem_limit=mem_limit, 
            output_file=os.path.join(cal_path, 'master_bias.fits'), 
            overwrite_output=True)
        
        imstatistics(master_bias, verbose=verbose)
        
        # Plot master bias
        plot2d(
            master_bias.data, title='master bias', show=show, save=save, path=fig_path)
        
        # Release memory
        del bias_list

    # Lamp concatenation
    if ('lamp.concatenate' in steps) or ('lamp' in steps):
        
        if verbose:
            print('\n[LAMP CONCATENATION]')
        
        # Load lamp
        if verbose:
            print('  - Loading lamp...')
        ifc_lamp = ImageFileCollection(
            location=data_dir, keywords=keywords, find_fits_by_reading=False, 
            filenames=list_dict['lamp'], glob_include=None, glob_exclude=None, ext=hdu)
        lamp_list = CCDDataList.read(
            file_list=ifc_lamp.files_filtered(include_path=True), hdu=hdu)
        
        # Trim
        if trim:
            if verbose:
                print('  - Trimming...')
            lamp_list = lamp_list.trim(row_range=row_range, col_range=col_range)
        
        lamp_list.statistics(verbose=verbose)
        
        # Subtract bias
        #   Uncertainties created here (equal to that of ``master_bias``) are useless!!!
        if verbose:
            print('  - Subtracting bias...')
        if 'master_bias' not in locals():
            master_bias = CCDData.read(os.path.join(cal_path, 'master_bias.fits'))
        lamp_list_bias_subtracted = lamp_list - master_bias
        
        lamp_list_bias_subtracted.statistics(verbose=verbose)
        
        # Create real uncertainty!!!
        if verbose:
            print('  - Creating deviation...')
        lamp_list_bias_subtracted_with_deviation = (
            lamp_list_bias_subtracted.create_deviation(
                gain=gain, readnoise=rdnoise, disregard_nan=True)
        )
        
        # Concatenate
        if verbose:
            print('  - Concatenating...')
        # Ensure that the first is the short exposure
        exptime = ifc_lamp.summary['exptime'].data
        if exptime[0] > exptime[1]:
            lamp_list_bias_subtracted_with_deviation = (
                lamp_list_bias_subtracted_with_deviation[::-1]
            )
        concatenated_lamp = concatenate(
            lamp_list_bias_subtracted_with_deviation, row_range=(0, None), 
            col_range=(0, index), scale=None)
        
        # Plot concatenated lamp
        plot2d(
            concatenated_lamp.data, title='concatenated lamp', show=show, save=save, 
            path=fig_path)
        
        # Write concatenated lamp to file
        concatenated_lamp.write(
            os.path.join(cal_path, 'concatenated_lamp.fits'), overwrite=True)
        
        # Release memory
        del (lamp_list, lamp_list_bias_subtracted, 
             lamp_list_bias_subtracted_with_deviation)
        
    # Curvature rectification
    if ('lamp.rectify' in steps) or ('lamp' in steps):
        
        if verbose:
            print('\n[CURVATURE RECTIFICATION]')
        
        # Fit coordinates
        if verbose:
            print('  - Fitting coordinates...')
        if 'concatenated_lamp' not in locals():
            concatenated_lamp = CCDData.read(
                os.path.join(cal_path, 'concatenated_lamp.fits'))
        U, _ = fitcoords(
            ccd=concatenated_lamp, slit_along=slit_along, order=1, n_med=15, n_piece=3, 
            prominence=1e-3, n_iter=3, sigma_lower=3, sigma_upper=3, grow=False, 
            use_mask=False, show=show, save=save, path=fig_path, height=0, threshold=0, 
            distance=5, width=5, wlen=15, rel_height=1, plateau_size=1)
        
        # Invert coordinate map
        if verbose:
            print('  - Inverting coordinate map...')
        X, Y = invertCoordinateMap(slit_along, U)
        np.save(os.path.join(cal_path, 'X.npy'), X)
        np.save(os.path.join(cal_path, 'Y.npy'), Y)
        
        # Rectify curvature
        if verbose:
            print('  - Rectifying curvature...')
        transformed_lamp = transform(ccd=concatenated_lamp, X=X, Y=Y)
        
        # Plot transformed lamp
        plot2d(
            transformed_lamp.data, title='transformed lamp', show=show, save=save, 
            path=fig_path)
        
        # Write transformed lamp to file
        transformed_lamp.write(
            os.path.join(cal_path, 'transformed_lamp.fits'), overwrite=True)
        
    # Wavelength calibration
    if ('lamp.calibrate' in steps) or ('lamp' in steps):
        
        if verbose:
            print('\n[WAVELENGTH CALIBRATION]')
        
        # Extract 1-dimensional spectrum
        if verbose:
            print('  - Extracting 1-dimensional lamp spectrum')
        if 'transformed_lamp' not in locals():
            transformed_lamp = CCDData.read(
                os.path.join(cal_path, 'transformed_lamp.fits'))
        lamp1d = extract(
            ccd=transformed_lamp, slit_along=slit_along, trace1d=750, n_aper=1, 
            aper_width=10, show=show, save=save, path=fig_path)

        # Correct dispersion
        if verbose:
            print('  - Correcting dispersion axis...')
        calibrated_lamp1d = dispcor(
            spectrum1d=lamp1d, reverse=True, reference=reference, n_piece=3, refit=True, 
            n_iter=5, sigma_lower=3, sigma_upper=3, grow=False, use_mask=True, show=show, 
            save=save, path=fig_path)

        # Plot calibrated lamp spectrum
        plotSpectrum1D(
            spectrum1d=calibrated_lamp1d, title='calibrated lamp', show=show, save=save, 
            path=fig_path)

        # Write calibrated lamp spectrum to file
        saveSpectrum1D(
            os.path.join(cal_path, 'lamp1d.fits'), calibrated_lamp1d, overwrite=True)
        
    # Flat combination
    if ('flat.combine' in steps) or ('flat' in steps):
        
        if verbose:
            print('\n[FLAT COMBINATION]')
            
        if verbose:
            print('  - Loading flat...')
        # Filter
        ifc_flat = ifc.filter(regex_match=True, imagetyp='Flat Field')
        # Load flat
        flat_list = CCDDataList.read(
            file_list=ifc_flat.files_filtered(include_path=True), hdu=hdu)
        
        # Trim
        if trim:
            if verbose:
                print('  - Trimming...')
            flat_list = flat_list.trim(row_range=row_range, col_range=col_range)
        
        flat_list.statistics(verbose=verbose)

        # Subtract bias
        #   Uncertainties created here (equal to that of ``master_bias``) are useless!!!
        if verbose:
            print('  - Subtracting bias...')
        if 'master_bias' not in locals():
            master_bias = CCDData.read(os.path.join(cal_path, 'master_bias.fits'))
        flat_list_bias_subtracted = flat_list - master_bias
        
        flat_list_bias_subtracted.statistics(verbose=verbose)

        # Combine flat
        #   Uncertainties created above are overwritten here!!!
        if verbose:
            print('  - Combining...')
        scaling_func = lambda ccd: 1 / np.ma.average(ccd)
        combined_flat = flat_list_bias_subtracted.combine(
            method='average', scale=scaling_func, sigma_clip=True, 
            sigma_clip_low_thresh=3, sigma_clip_high_thresh=3, 
            sigma_clip_func=np.ma.median, sigma_clip_dev_func=mad_std, 
            mem_limit=mem_limit, 
            output_file=os.path.join(cal_path, 'combined_flat.fits'), 
            overwrite_output=True)

        imstatistics(combined_flat, verbose=verbose)
        
        # Plot combined flat
        plot2d(
            combined_flat.data, title='combined_flat', show=show, save=save, 
            path=fig_path)
        
        # Release memory
        del flat_list, flat_list_bias_subtracted

    # Response
    if ('flat.normalize.response' in steps) or \
       ('flat.normalize' in steps) or \
       ('flat' in steps):
        
        if verbose:
            print('\n[RESPONSE]')
        
        # Response calibration
        if 'combined_flat' not in locals():
            combined_flat = CCDData.read(os.path.join(cal_path, 'combined_flat.fits'))
        reflat = response(
            ccd=combined_flat, slit_along=slit_along, n_piece=n_piece, n_iter=5, 
            sigma_lower=3, sigma_upper=3, grow=10, use_mask=True, show=show, save=save, 
            path=fig_path)

        imstatistics(reflat, verbose=verbose)
        
        # Plot response calibrated flat
        plot2d(reflat.data, title='reflat', show=show, save=save, path=fig_path)
        
        # Plot response mask
        plot2d(
            reflat.mask.astype(int), vmin=0, vmax=1, title='response mask', show=show, 
            save=save, path=fig_path)
        
        # Write response calibrated flat to file
        reflat.write(os.path.join(cal_path, 'reflat.fits'), overwrite=True)
    
    # Illumination
    if ('flat.normalize.illumination' in steps) or \
       ('flat.normalize' in steps) or \
       ('flat' in steps):

        if verbose:
            print('\n[ILLUMINATION]')
        
        # Illumination modeling
        if 'reflat' not in locals():
            reflat = CCDData.read(os.path.join(cal_path, 'reflat.fits'))
        reflat.mask |= custom_mask
        ilflat = illumination(
            ccd=reflat, slit_along=slit_along, method='Gaussian2D', sigma=sigma, 
            bins=10, n_iter=5, sigma_lower=3, sigma_upper=3, grow=5, use_mask=True, 
            show=show, save=save, path=fig_path)

        imstatistics(ilflat, verbose=verbose)

        # Plot illumination
        plot2d(ilflat.data, title='illumination', show=show, save=save, path=fig_path)
        
        # Plot illumination mask
        plot2d(
            ilflat.mask.astype(int), vmin=0, vmax=1, title='illumination mask', 
            show=show, save=save, path=fig_path)
        
        # Write illumination to file
        ilflat.write(os.path.join(cal_path, 'illumination.fits'), overwrite=True)

    # Flat normalization
    if ('flat.normalize' in steps) or ('flat' in steps):
        if verbose:
            print('\n[FLAT NORMALIZATION]')
        # Normalization
        normalized_flat = reflat.divide(ilflat, handle_meta='first_found')
        # Plot normalized flat
        plot2d(
            normalized_flat.data, title='normalized flat', show=show, save=save, 
            path=fig_path)
        # Write normalized flat to file
        normalized_flat.write(
            os.path.join(cal_path, 'normalized_flat.fits'), overwrite=True)
        # [can be replaced by a pre-defined custom mask]
        normalized_flat.mask = None

    # Correct targets
    if ('targ' in steps):
        
        if verbose:
            print('\n[TARGET CORRECTION]')
        
        # Load targ
        if verbose:
            print('  - Loading targ...')
        ifc_targ = ImageFileCollection(
            location=data_dir, keywords=keywords, find_fits_by_reading=False, 
            filenames=list_dict['targ'], glob_include=None, glob_exclude=None, ext=hdu)
        targ_list = CCDDataList.read(
            file_list=ifc_targ.files_filtered(include_path=True), hdu=hdu)
        
        # Trim
        if trim:
            if verbose:
                print('  - Trimming...')
            targ_list = targ_list.trim(row_range=row_range, col_range=col_range)

        targ_list.statistics(verbose=verbose)

        # Subtract bias
        #   Uncertainties created here (equal to that of ``master_bias``) are useless!!!
        if verbose:
            print('  - Subtracting bias...')
        if 'master_bias' not in locals():
            master_bias = CCDData.read(os.path.join(cal_path, 'master_bias.fits'))
        targ_list_bias_subtracted = targ_list - master_bias
        
        targ_list_bias_subtracted.statistics(verbose=verbose)

        # Create real uncertainty!!!
        if verbose:
            print('  - Creating deviation...')
        targ_list_bias_subtracted_with_deviation = (
            targ_list_bias_subtracted.create_deviation(
                gain=gain, readnoise=rdnoise, disregard_nan=True)
        )
        
        # Flat-fielding
        if verbose:
            print('  - Flat-fielding...')
        if 'normalized_flat' not in locals():
            normalized_flat = CCDData.read(os.path.join(cal_path, 'normalized_flat.fits'))
        targ_list_flat_fielded = (
            targ_list_bias_subtracted_with_deviation / normalized_flat
        )

        # Remove cosmic ray
        if verbose:
            print('  - Removing cosmic ray...')
        targ_list_cosmicray_corrected = targ_list_flat_fielded.cosmicray(
            method='Laplacian', use_mask=False, gain=gain, readnoise=rdnoise, sigclip=4.5, 
            sigfrac=0.3, objlim=1, niter=5, verbose=True)

        # Rectify curvature
        if verbose:
            print('  - Rectifying curvature...')
        if 'X' not in locals():
            X = np.load(os.path.join(cal_path, 'X.npy'))
        if 'Y' not in locals():
            Y = np.load(os.path.join(cal_path, 'Y.npy'))
        for i, targ_cosmicray_corrected in enumerate(targ_list_cosmicray_corrected):
            
            old_name = targ_cosmicray_corrected.header['FILENAME']
            new_name = f'{os.path.splitext(old_name)[0]}_corrected.fits'

            transformed_targ = transform(ccd=targ_cosmicray_corrected, X=X, Y=Y)

            # Plot transformed spectrum
            plot2d(
                transformed_targ.data, title=old_name, show=show, save=save, 
                path=fig_path)
            
            # Write transformed spectrum to file
            if verbose:
                print(f'  - Saving {new_name} to {pro_path}...')
            transformed_targ.write(
                os.path.join(pro_path, new_name), overwrite=True)


def main():
    """Command line tool."""
    
    # External parameters
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-i', '--data_dir', required=True, type=str, 
        help='Data (input) directory.'
    )
    parser.add_argument(
        '-s', '--semester', required=True, type=str, 
        help='Observation semester.'
    )
    parser.add_argument(
        '-w', '--slit_width', required=True, type=float, choices=[1.8, 2.3], 
        help='Slit width.'
    )
    parser.add_argument(
        '-o', '--save_dir', default='', type=str, 
        help='Saving (output) directory.'
    )
    parser.add_argument(
        '-r', '--reference', default=None, type=str, 
        help='Reference spectrum for wavelength calibration.'
    )
    parser.add_argument(
        '-m', '--mode', default='general', type=str, choices=['general', 'standard'], 
        help='General or standard.'
    )
        
    # Parse
    args = parser.parse_args()
    data_dir = os.path.abspath(args.data_dir)
    semester = args.semester
    slit_width = str(args.slit_width).replace('.', '')
    save_dir = os.path.abspath(args.save_dir)
    reference = args.reference
    mode = args.mode
    
    # Internal parameters
    hdu = 0
    shape = (2048, 2048)
    keywords = ['imagetyp', 'object', 'exptime']
    steps = ['header', 'trim', 'bias', 'lamp', 'flat', 'targ']
    # steps = ['trim', 'flat', 'targ']
    row_range = (329, 1830)
    col_range = (0, 1900)
    slit_along = 'col'
    index = 665
    n_piece = 23
    sigma = (20, 30)
    rdnoise = 4.64
    gain = 1.41
    
    semester_path = os.path.join(LIBRARY_PATH, semester)
    if not os.path.exists(semester_path):
        raise ValueError('Semester not found.')
    
    # Note that slit18 can be used to calibrate slit23
    if not reference:
        reference = os.path.join(
            semester_path, f'bfosc_g4_slit{slit_width}_{semester}.fits')
    else:
        reference = os.path.abspath(reference)
    if not os.path.exists(reference):
        raise ValueError('Reference not found.')
    
    region = os.path.join(semester_path, f'bfosc_g4_slit{slit_width}_{semester}.reg')
    custom_mask = getMask(region_name=region, shape=shape)
    
    pipeline(
        save_dir=save_dir, data_dir=data_dir, hdu=hdu, keywords=keywords, steps=steps, 
        row_range=row_range, col_range=col_range, slit_along=slit_along, 
        n_piece=n_piece, sigma=sigma, index=index, rdnoise=rdnoise, gain=gain, 
        custom_mask=custom_mask, reference=reference, mem_limit=500e6, show=False, 
        save=True, verbose=True, mode=mode)

if __name__ == '__main__':
    main()


"""
Data reduction pipelines for 2.16-m/BFOSC.
"""

__all__ = ['g4_old', 'phot_old', 'phot']

from .g4_old import pipeline as g4_old
from .phot_old import pipeline as phot_old
from .phot import pipeline as phot


"""
BFOSC photometry pipeline
"""

import os, argparse

# NumPy
import numpy as np
# AstroPy
from astropy.stats import mad_std
from astropy.nddata import CCDData
# ccdproc
from ccdproc import ImageFileCollection
# drpy
from drpy.batch import CCDDataList
from drpy.utils import imstatistics
from drpy.plotting import plot2d

from ..utils import makeDirectory, modifyHeader
from .utils import login, loadList

def pipeline(save_dir, data_dir, hdu, keywords, filters, steps, row_range, col_range, 
             rdnoise, gain, custom_mask, mem_limit, show, save, verbose, mode):
    
    """BFOSC photometry pipeline."""
    
    # Login message
    if verbose:
        login('photometry')
    
    # Make directories
    if verbose:
        print('\n[MAKE DIRECTORIES]')
        print(f'  - Changing working directory to {save_dir}...')
    os.chdir(save_dir)
    fig_path = makeDirectory(parent='', child='fig', verbose=verbose)
    pro_path = makeDirectory(parent='', child='pro', verbose=verbose)
    cal_path = makeDirectory(parent='', child='cal', verbose=verbose)

    # Load lists
    list_dict = dict()
    for f in filters:
        list_dict[f] = loadList(list_name=f)
        list_dict[f'flat_{f}'] = loadList(list_name=f'flat_{f}')
    
    # Construct image file collection
    ifc = ImageFileCollection(
        location=data_dir, keywords=keywords, find_fits_by_reading=False, 
        filenames=None, glob_include='*.fit', glob_exclude=None, ext=hdu)
    
    # Modify fits header
    #   Note that image file collection is constructed before header modification.
    if 'header' in steps:
        if verbose:
            print('\n[HEADER MODIFICATION]')
        for file_name in ifc.files_filtered(include_path=True):
            modifyHeader(file_name, verbose=verbose)
    
    if 'trim' in steps:
        custom_mask = custom_mask[row_range[0]:row_range[1], col_range[0]:col_range[1]]
        trim = True
    else:
        trim = False
    
    # Bias combination
    if ('bias.combine' in steps) or ('bias' in steps):
        
        if verbose:
            print('\n[BIAS COMBINATION]')
        
        # Load bias
        if verbose:
            print('  - Loading bias...')
        ifc_bias = ifc.filter(regex_match=True, imagetyp='Bias Frame')
        bias_list = CCDDataList.read(
            file_list=ifc_bias.files_filtered(include_path=True), hdu=hdu)
        
        # Trim
        if trim:
            if verbose:
                print('  - Trimming...')
            bias_list = bias_list.trim(row_range=row_range, col_range=col_range)
        
        bias_list.statistics(verbose=verbose)

        # Combine bias
        if verbose:
            print('  - Combining...')
        master_bias = bias_list.combine(
            method='average', sigma_clip=True, sigma_clip_low_thresh=3, 
            sigma_clip_high_thresh=3, sigma_clip_func=np.ma.median, 
            sigma_clip_dev_func=mad_std, mem_limit=mem_limit, 
            output_file=os.path.join(cal_path, 'master_bias.fits'), 
            overwrite_output=True)
        
        imstatistics(master_bias, verbose=verbose)
        
        # Plot master bias
        plot2d(
            master_bias.data, title='master bias', show=show, save=save, path=fig_path)
        
        # Release memory
        del bias_list
    
    # Flat combination
    if ('flat.combine' in steps) or ('flat' in steps):
        
        if verbose:
            print('\n[FLAT COMBINATION]')
        
        for f in filters:
            
            # Load flat
            if verbose:
                print(f'  - Loading {f} flat...')
            ifc_flat = ImageFileCollection(
                location=data_dir, keywords=keywords, find_fits_by_reading=False, 
                filenames=list_dict[f'flat_{f}'], glob_include=None, glob_exclude=None, 
                ext=hdu)
            flat_list = CCDDataList.read(
                file_list=ifc_flat.files_filtered(include_path=True), hdu=hdu)

            # Trim
            if trim:
                if verbose:
                    print('  - Trimming...')
                flat_list = flat_list.trim(row_range=row_range, col_range=col_range)

            flat_list.statistics(verbose=verbose)

            # Subtract bias
            #   Uncertainties created here (equal to that of ``master_bias``) are useless!!!
            if verbose:
                print('  - Subtracting bias...')
            if 'master_bias' not in locals():
                master_bias = CCDData.read(os.path.join(cal_path, 'master_bias.fits'))
            flat_list_bias_subtracted = flat_list - master_bias

            flat_list_bias_subtracted.statistics(verbose=verbose)

            # Combine flat
            #   Uncertainties created above are overwritten here!!!
            if verbose:
                print('  - Combining...')
            scaling_func = lambda ccd: 1 / np.ma.average(ccd)
            combined_flat = flat_list_bias_subtracted.combine(
                method='median', scale=scaling_func, sigma_clip=True, 
                sigma_clip_low_thresh=3, sigma_clip_high_thresh=3, 
                sigma_clip_func=np.ma.median, sigma_clip_dev_func=mad_std, 
                mem_limit=mem_limit, 
                output_file=os.path.join(cal_path, f'combined_flat_{f}.fits'), 
                overwrite_output=True)

            imstatistics(combined_flat, verbose=verbose)

            # Plot combined flat
            plot2d(
                combined_flat.data, title=f'combined_flat_{f}', show=show, save=save, 
                path=fig_path)

            # Release memory
            del flat_list, flat_list_bias_subtracted, combined_flat
    
    
    # Correct targets
    if ('targ' in steps):
        
        if verbose:
            print('\n[TARGET CORRECTION]')
        
        for f in filters:
            
            # Load targ
            if verbose:
                print(f'  - Loading {f} targ...')
            ifc_targ = ImageFileCollection(
                location=data_dir, keywords=keywords, find_fits_by_reading=False, 
                filenames=list_dict[f], glob_include=None, glob_exclude=None, ext=hdu)
            targ_list = CCDDataList.read(
                file_list=ifc_targ.files_filtered(include_path=True), hdu=hdu)

            # Trim
            if trim:
                if verbose:
                    print('  - Trimming...')
                targ_list = targ_list.trim(row_range=row_range, col_range=col_range)

            targ_list.statistics(verbose=verbose)

            # Subtract bias
            #   Uncertainties created here (equal to that of ``master_bias``) are useless!!!
            if verbose:
                print('  - Subtracting bias...')
            if 'master_bias' not in locals():
                master_bias = CCDData.read(os.path.join(cal_path, 'master_bias.fits'))
            targ_list_bias_subtracted = targ_list - master_bias

            targ_list_bias_subtracted.statistics(verbose=verbose)

            # Create real uncertainty!!!
            if verbose:
                print('  - Creating deviation...')
            targ_list_bias_subtracted_with_deviation = (
                targ_list_bias_subtracted.create_deviation(
                    gain=gain, readnoise=rdnoise, disregard_nan=True)
            )

            # Flat-fielding
            if verbose:
                print('  - Flat-fielding...')
            combined_flat = CCDData.read(
                os.path.join(cal_path, f'combined_flat_{f}.fits'))
            targ_list_flat_fielded = (
                targ_list_bias_subtracted_with_deviation / combined_flat
            )

            # Remove cosmic ray
            if verbose:
                print('  - Removing cosmic ray...')
            targ_list_cosmicray_corrected = targ_list_flat_fielded.cosmicray(
                method='Laplacian', use_mask=False, gain=gain, readnoise=rdnoise, sigclip=4.5, 
                sigfrac=0.3, objlim=1, niter=5, verbose=True)

            for i, targ_cosmicray_corrected in enumerate(targ_list_cosmicray_corrected):

                old_name = targ_cosmicray_corrected.header['FILENAME']
                new_name = f'{os.path.splitext(old_name)[0]}_{f}.fits'
                
                # Plot corrected target
                plot2d(
                    targ_cosmicray_corrected.data, title=old_name, show=show, 
                    save=save, path=fig_path)
                
                # Write corrected target to file
                if verbose:
                    print(f'  - Saving {new_name} to {pro_path}...')
                targ_cosmicray_corrected.write(
                    os.path.join(pro_path, new_name), overwrite=True)
    
    
def main():
    """Command line tool."""
    
    # External parameters
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-i', '--data_dir', required=True, type=str, 
        help='Data (input) directory.'
    )
    parser.add_argument(
        '-f', '--filters', required=True, type=str, 
        help='Filters.'
    )
    parser.add_argument(
        '-o', '--save_dir', default='', type=str, 
        help='Saving (output) directory.'
    )
    parser.add_argument(
        '-m', '--mode', default='general', type=str, choices=['general', 'standard'], 
        help='General or standard.'
    )
        
    # Parse
    args = parser.parse_args()
    data_dir = os.path.abspath(args.data_dir)
    filters = list(args.filters)
    save_dir = os.path.abspath(args.save_dir)
    mode = args.mode
    
    # Internal parameters
    hdu = 0
    shape = (2048, 2048)
    keywords = ['imagetyp', 'object', 'exptime']
    # steps = ['header', 'bias', 'flat', 'targ']
    steps = ['targ']
    row_range = None
    col_range = None
    rdnoise = 4.64
    gain = 1.41
    
    for f in filters:
        if f not in list('UBVRI'):
            raise ValueError(f'Filter {f} not available.')
    
    custom_mask = np.zeros(shape, dtype=bool)
    
    pipeline(
        save_dir=save_dir, data_dir=data_dir, hdu=hdu, keywords=keywords, 
        filters=filters, steps=steps, row_range=row_range, col_range=col_range, 
        rdnoise=rdnoise, gain=gain, custom_mask=custom_mask, mem_limit=500e6, 
        show=False, save=True, verbose=True, mode=mode)

    
if __name__ == '__main__':
    main()


import os
from glob import glob

# NumPy
import numpy as np
# AstroPy
from astropy.stats import mad_std
from astropy.table import Table
# ccdproc
from ccdproc import subtract_bias
# drpy
from drpy.batch import CCDDataList
from drpy.plotting import plot2d

    
def genFileTable(ccddatalist, keywords, verbose):
    """Generate file table."""
    
    if 'OBSTYPE' not in keywords:
        keywords.append('OBSTYPE')

    file_table = list()
    for ccd in ccddatalist:
        file_table.append([ccd.header[keyword] for keyword in keywords])

    file_table = Table(
        rows=file_table, names=[keyword for keyword in keywords])

    if verbose: file_table.pprint_all()
    
    return file_table


def loadData(data_dir, hdu, keywords, verbose):
    """Load and organize data"""

    file_list = sorted(glob(os.path.join(data_dir, '*.fits')))

    ccddatalist = CCDDataList.read(file_list=file_list, hdu=hdu)

    file_table = genFileTable(ccddatalist, keywords, verbose)

    mask_dark = file_table['OBSTYPE'] == 'dark'
    mask_flat = file_table['OBSTYPE'] == 'flat'
    mask_targ = file_table['OBSTYPE'] == 'object'

    return ccddatalist[mask_dark], ccddatalist[mask_flat], ccddatalist[mask_targ]


def getMasterFlat(flat_list, dark_list):
    """Master flat."""

    flat_with_lamp_on = flat_list.combine(
        method='average', sigma_clip=True, sigma_clip_low_thresh=3, 
        sigma_clip_high_thresh=3, sigma_clip_func=np.ma.median, 
        sigma_clip_dev_func=mad_std, mem_limit=500e6)

    flat_with_lamp_off = dark_list.combine(
        method='average', sigma_clip=True, sigma_clip_low_thresh=3, 
        sigma_clip_high_thresh=3, sigma_clip_func=np.ma.median, 
        sigma_clip_dev_func=mad_std, mem_limit=500e6)

    master_flat = subtract_bias(ccd=flat_with_lamp_on, master=flat_with_lamp_off)
    
    return master_flat


def main(data_dir, hdu, keywords, verbose):
    
    dark_list, flat_list, _ = loadData(
        data_dir=data_dir, hdu=hdu, keywords=keywords, verbose=verbose)

    # Assume that all the flats have the same configuration
    fsample, coadds, exptime = (
        flat_list[0].header['FSAMPLE'], flat_list[0].header['COADDS'], 
        flat_list[0].header['EXPTIME'])

    n_dark = len(dark_list)
    mask = np.zeros(n_dark, dtype=bool)
    for i in range(n_dark):
        if ((dark_list[i].header['FSAMPLE'] == fsample) & 
            (dark_list[i].header['COADDS'] == coadds) & 
            (dark_list[i].header['EXPTIME'] == exptime)):
            mask[i] = True

    # Mask flats with lamp off
    dark_list = dark_list[mask]

    for fccd, dccd in zip(flat_list, dark_list):
        print(fccd.header['FILENAME'], dccd.header['FILENAME'])

#     master_dark = dark_list.combine(
#         method='average', sigma_clip=True, sigma_clip_low_thresh=3, 
#         sigma_clip_high_thresh=3, sigma_clip_func=np.ma.median, 
#         sigma_clip_dev_func=mad_std, mem_limit=500e6)

#     master_dark.write('master_dark.fits', overwrite=True)

    master_flat = getMasterFlat(flat_list=flat_list, dark_list=dark_list)


    CCDData.writer.help('fits')

    master_flat.write('master_flat.fits', as_image_hdu=True, overwrite=True)

if __name__ == '__main__':

    data_dir = '/data3/zrn/data/Hale/TripleSpec/20221210'

    hdu = 0
    
    keywords = [
        'FILENAME', 'TIME', 'OBSTYPE', 'OBJECT', 'EXPTIME', 'COADDS', 'FSAMPLE', 
        'CATNAME', 'RA', 'DEC'
    ]
    
    verbose = True

    main(data_dir=data_dir, hdu=hdu, keywords=keywords, verbose=verbose)


import os, sys
from glob import glob

try:
    import numpy as np
except ImportError:
    print( 'Module `numpy` not found. Please install with: pip install numpy' )
    sys.exit()

try:
    from astropy import units as u
    from astropy.nddata import CCDData
except ImportError:
    print( 'Module `astropy` not found. Please install with: pip install astropy' )
    sys.exit()

try:
    import ccdproc
except ImportError:
    print( 'Module `ccdproc` not found. Please install with: pip install ccdproc' )
    sys.exit()

from utils import print_statistics, plot2d

def biascombine( path, ext, gain, rdnoise ):
    '''
    '''

    bias_frame_list = sorted( glob( os.path.join( path, '*.fit*' ) ) )

    biaslist = list()
    for i, bfile in enumerate( bias_frame_list ):
        # Load bias
        print( F'[Bias combination] Load {bfile} as bias frame [{i+1}/{len(bias_frame_list)}]' )
        bias = CCDData.read( bfile, hdu = ext, unit = u.adu )
        # Add uncertainty
        print( F'[Bias combination] Estimate uncertainty' )
        bias_with_deviation = ccdproc.create_deviation( bias, readnoise = rdnoise * u.adu )
        # Gain correction
        print( F'[Bias combination] Gain correction' )
        bias_gain_corrected = ccdproc.gain_correct( bias_with_deviation, gain * u.electron / u.adu )
        biaslist.append( bias_gain_corrected )

    print_statistics( biaslist )

    # Combine
    print( F'[Bias combination] Combine' )
    bias_combiner = ccdproc.Combiner( biaslist )
    master_bias = bias_combiner.median_combine()
    print_statistics( [master_bias] )

    # Plot
    print( F'[Bias combination] Plot' )
    plot2d( master_bias, 'Master Bias' )

    return master_bias

def flatcombine( filters, path, ext, expkw, gain, rdnoise, master_bias ):
    '''
    '''

    master_flat = dict()
    for filter in filters:

        flat_frame_list = sorted( glob( os.path.join( path, F'{filter}/flat/*.fit*' ) ) )

        flatlist = list(); expolist = list()
        for i, ffile in enumerate( flat_frame_list ):
            # Load flat-field
            print( F'[Flat-field combination] Load {ffile} as {filter} band flat-field [{i+1}/{len(flat_frame_list)}]' )
            flat = CCDData.read( ffile, hdu = ext, unit = u.adu )
            # Add uncertainty
            print( F'[Flat-field combination] Estimate uncertainty' )
            flat_with_deviation = ccdproc.create_deviation( flat, readnoise = rdnoise * u.adu )
            # Gain correction
            print( F'[Flat-field combination] Gain correction' )
            flat_gain_corrected = ccdproc.gain_correct( flat_with_deviation, gain * u.electron / u.adu )
            # Bias subtraction
            print( F'[Flat-field combination] Bias subtraction' )
            flat_bias_subtracted = ccdproc.subtract_bias( flat_gain_corrected, master_bias )
            flatlist.append( flat_bias_subtracted ); expolist.append( flat_bias_subtracted.meta[expkw] )

        print_statistics( flatlist )

        # Combine
        print( F'[Flat-field combination] Combine' )
        flat_combiner = ccdproc.Combiner( flatlist )
        flat_combiner.scaling = 1 / np.array( expolist )
        master_flat[filter] = flat_combiner.median_combine()
        print_statistics( [master_flat[filter]] )

        # Plot
        print( F'[Flat-field combination] Plot' )
        plot2d( master_flat[filter], F'{filter} Band Master Flat-field' )

    return master_flat

def imagecorrection( filters, path, ext, gain, rdnoise, master_bias, master_flat ):
    '''
    '''

    images = dict()
    for filter in filters:

        img_frame_list = sorted( glob( os.path.join( path, F'{filter}/*.fit*' ) ) )

        images[filter] = list()
        for i, imgfile in enumerate( img_frame_list ):
            # Load object
            print( F'[Image correction] Load {imgfile} as {filter} band object image [{i+1}/{len(img_frame_list)}]' )
            img = CCDData.read( imgfile, hdu = ext, unit = u.adu )
            # Add uncertainty
            print( F'[Image correction] Estimate uncertainty' )
            img_with_deviation = ccdproc.create_deviation( img, readnoise = rdnoise * u.adu )
            # Gain correction
            print( F'[Image correction] Gain correction' )
            img_gain_corrected = ccdproc.gain_correct( img_with_deviation, gain * u.electron / u.adu )
            # Bias subtraction
            print( F'[Image correction] Bias subtraction' )
            img_bias_subtracted = ccdproc.subtract_bias( img_gain_corrected, master_bias )
            # Flat-fielding
            print( F'[Image correction] Flat-fielding' )
            img_flat_fielded = ccdproc.flat_correct( img_bias_subtracted, master_flat[filter] )
            # Cosmic ray correction
            print( F'[Image correction] Mask cosmic ray pixels' )
            reduced_image = ccdproc.cosmicray_lacosmic( img_flat_fielded,
                                                        objlim    = 1,
                                                        sigclip   = 4.5, 
                                                        sigfrac   = 0.5,
                                                        readnoise = gain * rdnoise * u.electron,
                                                        verbose   = False )
            images[filter].append( reduced_image )

        print_statistics( images[filter] )
    
    return images